"""Common utilities used by the module."""

import importlib
from typing import TYPE_CHECKING, Any, NamedTuple, Optional, Type, Union, overload

from django.conf import settings
from django.core.cache import caches
from django.core.cache.backends.base import BaseCache
from django.core.exceptions import ValidationError

from .constants import CACHE_KEY_PREFIX
from .serializers import BaseSerializer


if TYPE_CHECKING:
    from typing_extensions import Literal

    from .models import Config


_SENTINEL = object()


class DataTuple(NamedTuple):
    """
    Data coming from cache.

    Optional representation when you're not sure if you stored
    a None in the cache, or nothing is found.
    """

    value: Any
    allow_template_use: bool


def _cache_key_transform(key: str) -> str:
    """Transforms the configuration key to avoid undesired collisions in cache."""

    return CACHE_KEY_PREFIX + key.lower()


def _get_cache() -> BaseCache:
    """Return the configured cache backend."""

    return caches[settings.ABOUTCONFIG_CACHE_NAME]


def _set_cache(config: "Config") -> None:
    """Put the given configuration data into cache."""

    cache = _get_cache()
    cache_key = _cache_key_transform(config.key)
    cache.set(
        cache_key,
        DataTuple(config.get_value(), config.allow_template_use),
        settings.ABOUTCONFIG_CACHE_TTL,
    )


def _delete_cache(config: "Config") -> None:
    """Remove value from cache."""

    cache = _get_cache()
    cache_key = _cache_key_transform(config.key)
    cache.delete(cache_key)


def load_class(class_path: str) -> Type:
    """Load class from absolute class path."""

    split_path = class_path.split(".")
    class_name = split_path.pop()
    module_path = ".".join(split_path)

    module = importlib.import_module(module_path)
    return getattr(module, class_name)


def load_serializer(class_path: str) -> Type[BaseSerializer]:
    """Load a class by name/path that implements the serialization interface.

    Returns an instance of the class or raises ImportError/ValueError/AttributeError."""

    klass = load_class(class_path)

    if not BaseSerializer.is_class_valid(klass):
        raise ValueError('"{}" is not a valid serializer'.format(class_path))

    return klass


def serializer_validator(class_path: str) -> None:
    """Check whether the given class (via path) is a valid serialzier or not.

    Raises ValidationError if not valid."""

    try:
        load_serializer(class_path)
    except (ValueError, ImportError, AttributeError):
        raise ValidationError("Invalid serializer class")


# pylint: disable=missing-function-docstring
@overload
def get_config(key: str, value_only: "Literal[True]") -> Optional[Any]:
    ...


@overload
def get_config(key: str, value_only: "Literal[False]") -> DataTuple:
    ...


@overload
def get_config(key: str, value_only: bool) -> Union[DataTuple, Optional[Any]]:
    ...


def get_config(key: str, value_only: bool = True) -> Union[DataTuple, Optional[Any]]:
    """
    Get the configuration value for the given key.

    By default this function only returns the value, set value_only=False to get a
    2-tuple with the following elements: (value, allow_template_use). This functionality
    is almost certainly not needed by end-users, only included for the built-in template
    tag filtering functionality.

    Returns None if no such configuration exists (or (None, True) if value_only=False).
    """

    from .models import Config  # pylint: disable=import-outside-toplevel,cyclic-import

    cache = _get_cache()
    cache_key = _cache_key_transform(key)
    data = cache.get(cache_key, _SENTINEL)

    if data is _SENTINEL:
        try:
            config = Config.objects.select_related("data_type").get(key=key.lower())
        except Config.DoesNotExist:
            data = DataTuple(None, True)
            cache.set(cache_key, data, settings.ABOUTCONFIG_CACHE_TTL)
        else:
            data = DataTuple(config.get_value(), config.allow_template_use)
            _set_cache(config)

    return data.value if value_only else data


def preload_cache() -> None:
    """Load all configuration data into cache."""

    from .models import Config  # pylint: disable=import-outside-toplevel,cyclic-import

    for config in Config.objects.all():
        _set_cache(config)



"""User configuration access via the django admin."""

from typing import TYPE_CHECKING, List, Tuple

from django.conf.urls import url
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.http import Http404, HttpRequest, HttpResponse, HttpResponseBadRequest, JsonResponse
from django.urls.resolvers import RegexPattern

from .constants import CONFIG_ADMIN_TYPE_QUERY_KEY
from .forms import ConfigAdminForm
from .models import Config, DataType


if TYPE_CHECKING:
    from django.db.models import QuerySet


@admin.register(DataType)
class DataTypeAdmin(admin.ModelAdmin):
    """Django admin for the DataType model."""

    list_display = ("name", "serializer_class")


class ConfigNamespaceFilter(admin.SimpleListFilter):
    """Key namespace sidebar filter implementation."""

    title = "namespace"
    parameter_name = "namespace"

    def lookups(self, request: HttpRequest, model_admin: "ConfigAdmin") -> List[Tuple[str, str]]:
        queryset = model_admin.get_queryset(request)
        lookups = set(queryset.values_list("key_namespace", flat=True).distinct().iterator())

        return sorted((x, x) for x in lookups)

    def queryset(self, request: HttpRequest, queryset: "QuerySet[Config]") -> "QuerySet[Config]":
        if self.value() is None:
            return queryset

        return queryset.filter(key_namespace=self.value())


@admin.register(Config)
class ConfigAdmin(admin.ModelAdmin):
    """Django admin for the Config model."""

    list_display = ("key", "value", "default_value", "data_type", "allow_template_use", "in_cache")
    list_filter = ("data_type", "allow_template_use", ConfigNamespaceFilter)
    search_fields = ("key",)
    form = ConfigAdminForm

    class Media:
        js = ("aboutconfig/config_admin.js",)

    @staticmethod
    def fetch_field_view(request: HttpRequest) -> HttpResponse:
        """Fetches the HTML required for rendering the given type of value."""

        try:
            data_type = DataType.objects.get(pk=request.GET[CONFIG_ADMIN_TYPE_QUERY_KEY])
        except (DataType.DoesNotExist, ValueError, TypeError, KeyError):
            raise Http404()

        try:
            attrs = {"id": request.GET["id"]}
        except KeyError:
            return HttpResponseBadRequest()

        current_value = request.GET.get("value", "")

        try:
            config = Config.objects.get(pk=request.GET["config_pk"])
        except (DataType.DoesNotExist, ValueError, TypeError, KeyError):
            config = Config(key="x", key_namespace="x")

        config.data_type = data_type
        config.value = current_value

        try:
            config.full_clean()
        except ValidationError:
            current_value = ""

        widget = data_type.get_widget_class()(**data_type.widget_args)
        return JsonResponse({"content": widget.render("value", current_value, attrs=attrs)})

    def get_urls(self) -> List[RegexPattern]:
        """Get custom admin endpoints."""

        return [
            url(
                r"^fetch-value-field-html",
                self.admin_site.admin_view(self.fetch_field_view),
                name="ac-fetch-value-field",
            )
        ] + super().get_urls()



"""Provides ability to (re)load config cache from the command line."""

from typing import Any

from django.core.management.base import BaseCommand

from aboutconfig import utils


class Command(BaseCommand):
    """Loads all existing config entries into cache"""

    help = "Loads all existing config entries into cache"

    def handle(self, *args: Any, **options: Any) -> None:
        utils.preload_cache()



"""Forms for model editing."""

from typing import Any

from django import forms
from django.urls import reverse

from .models import Config, DataType


class ConfigAdminForm(forms.ModelForm):
    """Form used for editing Config objects in the django admin."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Constructor.

        Replaces the default widgets with new ones."""

        super().__init__(*args, **kwargs)

        self.fields["data_type"].widget = forms.Select(
            choices=self.fields["data_type"].widget.choices,
            attrs={
                "class": "aboutconfig-type-field",
                "data-url": reverse("admin:ac-fetch-value-field"),
                "data-instance-pk": self.instance.pk or "",
            },
        )

        data_type = self.get_data_type()

        original_field = self.fields["value"]
        field_class = original_field.__class__
        self.fields["value"] = field_class(
            initial=original_field.initial,
            widget=data_type.get_widget_class()(**data_type.widget_args),
        )

    class Meta:
        model = Config
        fields = ("data_type", "key", "value", "allow_template_use")
        readonly_fields = ("default_value",)

    def get_data_type(self) -> DataType:
        """Get data type to use when figuring out how to render the value field."""

        try:
            return DataType.objects.get(pk=self.data["data_type"])
        except (KeyError, DataType.DoesNotExist):
            pass

        try:
            return self.instance.data_type
        except DataType.DoesNotExist:
            return DataType.objects.get(name="String")



"""A firefox-like about:config implementation for one-off settings in Django apps."""

from typing import Any


__version__ = (1, 1, 0)
default_app_config = "aboutconfig.apps.AboutconfigConfig"  # pylint: disable=invalid-name


def get_config(key: str, value_only: bool = True) -> Any:
    """Get configured value by key.

    By default returns value only. If ``value_only`` is ``False``, returns an instance of
    aboutconfig.utils.DataTuple which also contains the ``allow_template_use`` value.

    This is a lazy wrapper around the internal ``utils.get_config()`` function."""

    from . import utils  # pylint: disable=import-outside-toplevel

    return utils.get_config(key, value_only)



# -*- coding: utf-8 -*-

import os

from setuptools import find_packages, setup

from aboutconfig import __version__


BASE_DIR = os.path.dirname(os.path.abspath(__file__))
VERSION_STRING = ".".join(str(s) for s in __version__)

with open(os.path.join(BASE_DIR, "README.rst")) as fp:
    long_description = fp.read()

setup(
    name="django-aboutconfig",
    version=VERSION_STRING,
    url="https://gitlab.com/impala1/django-aboutconfig",
    license="GPLv3+",
    description="A firefox-like about:config implementation for one-off settings in Django apps.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    keywords=["django", "library", "configuration"],
    author="Kirill Stepanov",
    author_email="mail@kirillstepanov.me",
    packages=find_packages(),
    download_url=(
        "https://gitlab.com/impala1/django-aboutconfig/-/archive/{version}/"
        "django-aboutconfig-{version}.tar.gz".format(version=VERSION_STRING)
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=["Django>=2.0",],
    python_requires=">=3.6",
    include_package_data=True,
    zip_safe=False,
)
